#!/usr/bin/env python3
"""
This example illustrates how to interpolate 1-d distributions
nonparametrically as a function of a single predictor
"""

__author__="Igor Volobouev (i.volobouev@ttu.edu)"
__version__="1.0"
__date__ ="Oct 10 2019"

# Perform necessary imports
import npstat as ns
import matplotlib.pyplot as plt

# Interpolation grid used to connect parameter values with
# distributions. We will only interpolate two distributions
# assuming that one of them corresponds to parameter value
# 0.0 and the other corresponds to 1.0.
paramGrid = ns.GridAxis(0.0, 1.0)

# Create the object which will perform density interpolation.
# We can interpolate vertically or we can interpolate the quantile
# function (horizontally). Change the "interpolateVertically"
# setting below and see how the results change.
interpolateVertically = False
interpolator = ns.InterpolatedDistro1D1P(paramGrid, interpolateVertically)

# Assign distributions to each cell of the interpolation grid
interpolator.setGridDistro(0, ns.Gauss1D(0.0, 0.5))
interpolator.setGridDistro(1, ns.Uniform1D(0.0, 1.0))

# Cycle over some parameter values and plot interpolated distributions
paramValues = ns.EquidistantInLinearSpace(0.0, 1.0, 9)
fig = plt.figure()
for i, param in enumerate(paramValues):
    interpolator.setParameter(param)
    x, y = ns.scanDensity1D(interpolator, -2.0, 2.0, 400)
    ax = fig.add_subplot(3, 3, i+1)
    ax.plot(x, y)
    ax.set_ylim(-0.2, 1.2)
    ax.set_xlabel('X')
    ax.set_ylabel('Density')
    ax.set_title("p = {:.3f}".format(param))

# Display the window
fig.suptitle("Interpolating Between Gaussian and Uniform")
fig.tight_layout(rect=[0, 0, 1, 0.95])
plt.show()
