#!/usr/bin/env python3
"""
This example illustrates the NPStat code that can be used to solve
one-dimensional heat equation with Neumann boundary conditions (no heat
transfer through the boundaries) and variable thermal diffusivity. The
Green's function (GF) of such an equation is a doubly stochastic matrix,
as it transforms a uniform initial temperature distribution into uniform
and also conserves energy. Therefore, this equation can be used to
generate variable bandwidth doubly stochastic filters useful for copula
smoothing.
"""

__author__="Igor Volobouev (i.volobouev@ttu.edu)"
__version__="1.0"
__date__ ="Jan 31 2020"

# Perform necessary imports
import npstat as ns
import numpy as np
import matplotlib.pyplot as plt
from scipy import interpolate
from npstat_utils import BoxND
from npstat_plotting import colormeshNumpyArray

# Some parameters for the script
nDiscrete = 51             # Number of spatial discretization points.
spatialStep = 1.0          # Spatial step size.
boundaryDiffusivity = 5.0  # Thermal diffusivity at the boundaries of
                           # the interval.
midpointDiffusivity = 1.0  # Thermal diffusivity at the center of the
                           # interval. The diffusivity will be modelled
                           # with a second order polynomial.
maxTimeSteps = 200         # Maximum number of time steps to make. For
                           # a flat diffusivity profile, the typical GF
                           # width away from the boundary is expected
                           # to be sqrt(2*diffusivity*t).

# Build the thermal diffusivity function
nIntervals = nDiscrete - 1
xmax = nIntervals*spatialStep
thD = (boundaryDiffusivity, midpointDiffusivity, boundaryDiffusivity)
thDFunction = interpolate.interp1d((0.0, xmax/2.0, xmax), thD, 'quadratic')

# Collection of thermal diffusivity values along the interval
xCoords = np.linspace(0.0, xmax, nDiscrete)
thDValues = thDFunction(xCoords)

# Even though the heat equation is solved using the unconditionally
# stable Crank-Nicolson scheme, a high fidelity Green's function needs
# a small time step, like in the FTCS scheme.
maxDiffusivity = max(boundaryDiffusivity, midpointDiffusivity)
dt = spatialStep*spatialStep/2.0/maxDiffusivity

# The heat equation solver
eq = ns.HeatEq1DNeumannBoundary(thDValues, spatialStep, dt)

# Prepare to plot the Green's function
box = BoxND((0.0, xmax), (0.0, xmax))
fig = plt.figure(figsize=(10, 5))
ax1 = fig.add_subplot(121)

# Prepare to plot effective bandwidth (standard deviation of the GF)
ax2 = fig.add_subplot(122)
ax2.set_xlabel('X')
ax2.set_ylabel('h')
ax2.set_title("Effective Bandwidth")
line1, = ax2.plot(xCoords, np.sqrt(2*dt*thDValues), 'b',
                  label="Naive Expectation")
line2, = ax2.plot(xCoords, np.sqrt(2*dt*thDValues), 'r',
                  label="Actual GF Width")
ax2.legend(loc=9)

plt.draw()

# Turn interactive mode on for dynamic updates
plt.ion()

# Run the solver updating the plot
for counter in range(maxTimeSteps):
    # Make a time step and get the updated GF
    eq.step()
    t = eq.time()
    gf = ns.matrixToNumpy(eq.GF())

    # Update the GF plot
    colormeshNumpyArray(ax1, gf, box)
    ax1.set_title("Heat Equation Green's Function, t = {:.1f}".format(t))

    # Update the effective bandwidth values
    line1.set_ydata(np.sqrt(2*t*thDValues))
    line2.set_ydata([eq.gfWidth(cell) for cell in range(nDiscrete)])
    ax2.relim()
    ax2.autoscale_view(True, False, True)

    # Give the windowing system some time to redraw the plots
    plt.pause(0.01)

# Turn off the interactive mode
plt.ioff()
plt.show()
