#!/usr/bin/env python3
"""
This example illustrates how to save various NPStat and python objects
in Geners archives. Geners distribution itself comes with an extensive
set of C++ examples explaining how to use these archives. All of that
functionality is also accessible through the Python API.

The stored objects can be retrieved from the archives by both Python and
C++ code. The only exception are the native python objects stored in the
pickled form. In principle, these objects can be unpickled in C++, but
they do not have a C++ API and, therefore, they are not terribly useful
in C++ programs.
"""

__author__="Igor Volobouev (i.volobouev@ttu.edu)"
__version__="1.0"
__date__ ="November 16 2019"

# Perform necessary imports
import npstat as ns
from npstat_utils import HistoND
from npstat_utils import Record, CPPRecord, Reference


# The following function will be used to fill a Geners archive with items
def store_items(ar):
    # Make an example distribution
    d1 = ns.Gauss1D(0, 1)

    # Store the example distribution in the archive. Archives store
    # "records". Each record is comprised of the object itself,
    # object name (second argument of the "Record" function), and object
    # category (third argument of the "Record" function). Names and
    # categories are just user-defined strings, subsequently used for
    # searching and retrieving stored items from the archive.
    ar.store(Record(d1, "d1", "Distributions"))

    # Make an example histogram
    histoAxis = ns.HistoAxis(20, -5, 5, "X")
    h1 = HistoND("int", histoAxis, "Example NPStat Histogram", "Counts")
    rng = ns.MersenneTwister()
    ns.randomHistoFill1D(h1, rng, d1, 1000)

    # Store the example histogram
    ar.store(Record(h1, "h1", "Histograms"))

    # Make a native python object
    mylist = [1, "abc", 2.0]

    # Store the python object. By default, native python objects are
    # stored in the pickled form.
    ar.store(Record(mylist, "mylist", "PickledObjects"))

    # Certain python types that have direct C++ analogs can be stored
    # in a C++ form. These are numbers, strings, and homogeneous lists
    # of numbers and strings. Objects stored in this manner can be later
    # retrieved and used from C++ code. For more details, see help on
    # function "CPPRecord" in module npstat_utils. Compared to "Record",
    # the CPPRecord function requires one more argument (the first one)
    # which is the C++ type into which the python type is converted.
    intNumber = 7
    floatNumber = 3.21
    floatSeq = [5.0, 4.0, 3.0, 2.0, 1.0]
    someName = "John Doe"
    ar.store(CPPRecord("int", intNumber, "intNumber", "CPPObjects"))
    ar.store(CPPRecord("double", floatNumber, "floatNumber", "CPPObjects"))
    ar.store(CPPRecord("vector<double>", floatSeq, "floatSeq", "CPPObjects"))
    ar.store(CPPRecord("string", someName, "someName", "CPPObjects"))


# The following function illustrates retrieval of the stored items from
# the archive. Naturally, the code of this function must be synchronized
# with the code of "store_items". Normally, item storage and retrieval
# will be performed by different scripts run at different times.
def retrieve_items(ar):
    # Each reference can return more than one item. However, below we use
    # unique specifications of the item name and category, and expect only
    # one item to be returned.
    d1 = Reference(ar, "d1", "Distributions").retrieve(0)
    h1 = Reference(ar, "h1", "Histograms").retrieve(0)
    mylist = Reference(ar, "mylist", "PickledObjects").retrieve(0)

    # C++ analogs of native python types should be retrieved by value.
    intNumber = Reference(ar, "intNumber", "CPPObjects").getValue(0)
    floatNumber = Reference(ar, "floatNumber", "CPPObjects").getValue(0)
    floatSeq = Reference(ar, "floatSeq", "CPPObjects").getValue(0)
    someName = Reference(ar, "someName", "CPPObjects").getValue(0)

    print("Retrieved d1, {}({}, {})".format(
        d1.classId().name(), d1.location(), d1.scale()))
    print('Retrieved h1, {} with title "{}"'.format(
        h1.classId().name(), h1.title()))
    print("Retrieved mylist =", mylist)
    print("Retrieved intNumber =", intNumber)
    print("Retrieved floatNumber =", floatNumber)
    print("Retrieved floatSeq =", floatSeq)
    print('Retrieved someName = "{}"'.format(someName))


# Function for printing catalog entries
def print_catalog_entry(entry):
    id = entry.id()
    className = entry.type().name()
    itemName = entry.name()
    itemCategory = entry.category()
    print("{:<3} {:42} {:15} {:15}".format(
        id, className, itemName, itemCategory))


# The following function will be used to examine the archive catalog
def examine_catalog(ar):
    print('\nArchive "{}" contains {} items:'.format(ar.name(), ar.size()))
    # Unique item ids in the archive are just integers. The
    # archive methods "smallestId" and "largestId" retrieve
    # the smallest and the largest item id. So, the iteration over the
    # item descriptions in the catalog can be performed as follows:
    for itemNum in range(ar.smallestId(), ar.largestId()+1):
        if ar.itemExists(itemNum):
            print_catalog_entry(ar.getCatalogEntry(itemNum))


# Create a string archive. This archive resides in memory.
# It is fast, but its size is limited by the available memory.
strAr = ns.StringArchive("Example string archive")
store_items(strAr)

# String archives can be written to disk and read back using the
# following pairs of functions: writeStringArchive/readStringArchive
# or writeCompressedStringArchive/readCompressedStringArchive.
# We will use an uncompressed version here. File extension ".gssa"
# stands for "generic serialization string archive".
strArchiveFile = "persistence.gssa"
ns.writeStringArchive(strAr, strArchiveFile)

# We no longer need the original string archive, so we might as well
# release some memory
del strAr

# Read back the string archive from disk
strArRead = ns.readStringArchive(strArchiveFile)

# Print some info from the read back archive
examine_catalog(strArRead)

# Retrieve items from the archive
retrieve_items(strArRead)
del strArRead

# Perform similar operations with a file archive. This archive can grow
# as large as the file system allows. It is closed and fully written to
# disk when its destructor is called.
archiveName = "persistence"
writeAr = ns.BinaryFileArchive(archiveName, "w")
store_items(writeAr)
del writeAr

# Read the file archive back from disk
readAr = ns.BinaryFileArchive(archiveName, "r")
examine_catalog(readAr)
retrieve_items(readAr)

# At this point, you can also examine the archive catalogs with Geners
# tools "cdump" and "gssa_dump". See https://geners.hepforge.org/tools.html
# for more details.
