#!/usr/bin/env python3
"""
This variation of plot_histo_1d_matplotlib.py script shows how to change
histogram orientation, add a hatching pattern, and change the fill color
"""

__author__="Igor Volobouev (i.volobouev@ttu.edu)"
__version__="1.0"
__date__ ="Sep 13 2019"

# Perform necessary imports
import npstat as ns
from npstat_utils import HistoND
import matplotlib.pyplot as plt
import matplotlib
import numpy as np

# Create the histogram axis (needed to construct the histogram below).
# Here, we use an axis with uniform binning. The axis constructor arguments
# are the number of bins, left histogram edge, right histogram edge, and
# the axis label.
histoAxis = ns.HistoAxis(20, -5.0, 5.0, "X")

# The following line creates a 1-d histogram with bins of type "int"
h1 = HistoND("int", histoAxis, "Example NPStat Histogram", "Counts")

# We will fill the histogram with Gaussian random numbers. In order to do
# that, we need a random number generator and a statistical distribution
# object which will be used to convert random numbers on [0, 1) into the
# random numbers distributed as we like. The default constructor of
# MersenneTwister generator picks up a different random seed every time,
# so that the histogram will be different every time you run this script.
rng = ns.MersenneTwister()
d1 = ns.Gauss1D(0, 1)

# Fill the histogram with 1000 random numbers
h1.fill(d1.generate(rng, 1000), 1)

# Create Matplotlib plot window
fig = plt.figure()
ax = fig.add_subplot(111)

# Generate xdata, ydata sequence for the outline drawn for 1-d histograms
x, y = ns.histoOutline1D(h1)

# The "plot_histo_1d.py" script employed the statement "ax.fill(x, y)".
# Here, we are changing the ordering of the axes, adding a hatch,
# and changing the color.
ax.add_artist(matplotlib.patches.Polygon(np.column_stack((y,x)),
                                         hatch='x', facecolor='green'))

# Set the x and y ranges of the plot (x and y are reversed in comparison
# with "plot_histo_1d.py")
ax.set_ylim(h1.axis(0).min(), h1.axis(0).max())
ax.set_xlim(0, 1.1*max(y))

# Set the axis labels and the plot title
plt.ylabel(h1.axis(0).label())
plt.xlabel(h1.accumulatedDataLabel())
plt.title(h1.title())

# Display the window. The code enters Matplotlib event loop here.
# You will be able to use the window interface to manipulate the
# figure (pan, zoom) and to save it in various graphical formats.
plt.show()
