#!/usr/bin/env python3
"""
This example illustrates how to fill NPStat 2-d histograms with random
number and how to plot them using "pcolormesh" Matplotlib facility
"""

__author__="Igor Volobouev (i.volobouev@ttu.edu)"
__version__="1.0"
__date__ ="Sep 29 2019"

# Perform necessary imports
import npstat as ns
from npstat_utils import HistoND
import matplotlib.pyplot as plt
from npstat_plotting import colormeshHisto2D

# To construct an n-dimensional histogram, we need to provide
# a collection of n axes as the constructor argument.
xaxis = ns.HistoAxis(50, -5.0, 5.0, "X")
yaxis = ns.HistoAxis(25, -2.0, 3.0, "Y")

# The following line creates a 2-d histogram with bins of type "double".
# The histogram dimensionality is inferred from the number of axes provided.
h2 = HistoND("int", (xaxis, yaxis), "Example 2-d Histogram", "Z")

# We will fill the histogram with random numbers, so we need a generator
rng = ns.MersenneTwister()

# The flat random numbers on [0, 1) will be converted into multivariate
# Gaussian random numbers using an object of NPStat "GaussND" class.
center = (1.0, 0.5)
covarianceMatrix = ((2.0, 0.3),
                    (0.3, 0.25))
distro = ns.GaussND(center, covarianceMatrix)

# Fill the histogram. The fastest way to fill a histogram with
# random numbers is provided by the NPStat utility functions
# "randomHistoFillND" and "randomHistoFillND". You can use these
# functions if you don't need the random sample itself for any
# other purpose.
ns.randomHistoFillND(h2, rng, distro, 10000)

# Prepare Matplotlib figure for plotting the color map
fig = plt.figure()
axes = fig.add_subplot(111)

# Plot the histogram using the default color map
mesh = colormeshHisto2D(axes, h2, alpha=0.9)

# Add the color bar
cbar = fig.colorbar(mesh)
cbar.ax.set_title(h2.accumulatedDataLabel())

# Display the window. The code enters Matplotlib event loop here.
# You will be able to use the window interface to manipulate the figure
# (rotate, pan, zoom) and to save it in various graphical formats.
plt.show()
