#ifndef NPSTAT_ABSBANDWIDTHCV_HH_
#define NPSTAT_ABSBANDWIDTHCV_HH_

/*!
// \file AbsBandwidthCV.hh 
//
// \brief Interface definitions for KDE or LOrPE cross-validation calculations
//
// Implementations of these interfaces should calculate the quantity
// which is to be maximized (pseudo log likelihood, -AMISE, etc.)
//
// Author: I. Volobouev
//
// September 2010
*/

#include <vector>
#include <utility>

#include "npstat/stat/HistoND.hh"

#ifdef SWIG
#include <stdexcept>
#include "npstat/wrap/arrayNDToNumpy.hh"
#include "npstat/wrap/numpyArrayUtils.hh"
#endif // SWIG

namespace npstat {
    // Forward declarations
    struct AbsPolyFilter1D;
    struct AbsPolyFilterND;

    /**
    // Cross-validation for univariate densities.
    // "Numeric" template parameter is the type of histogram bins.
    // "Num2" is the type produced by the density estimation code
    // (usually just double). "Num3" is the type of point coordinates
    // used to fill the histogram (normally, float or double), and
    // "Num4" is the type of point weights.
    */
    template<
        typename Numeric,
        typename Num2,
        typename Num3 = double,
        typename Num4 = double
    >
    struct AbsBandwidthCV1D
    {
        typedef Numeric bin_type;
        typedef Num2 density_type;
        typedef Num3 coord_type;
        typedef Num4 weight_type;

        inline virtual ~AbsBandwidthCV1D() {}

        /**
        // It should be assumed that the "nFillsInRange" method of the
        // argument histogram returns the actual number of fills (that is,
        // the histogram represents an actual collection of points, has
        // possible bin values of 0, 1, 2, ..., and it is not scaled).
        //
        // "densityEstimate" is allowed to be an estimate without truncation
        // (even if it includes negative values). Dependence of the
        // optimized quantity on the bandwidth should be smoother without
        // truncation. Naturally, the order of values in "densityEstimate"
        // is the same as the order of bins in "histo".
        //
        // The "filterUsed" parameter is needed in order to be able to
        // do "leave-one-out" type of cross-validation. In such
        // calculations one has to know how much of the density estimate 
        // at coordinate x is contributed by the source point at x. When
        // density estimates are done on the grid, the filters differ for
        // the center of the grid and on the border. Because of this, we
        // will need to have an access to the complete filter collection.
        */
        virtual double operator()(
            const HistoND<Numeric>& histo,
            const Num2* densityEstimate, unsigned lenEstimate,
            const AbsPolyFilter1D& filterUsed) const = 0;

        /** Cross-validation for samples of univariate weighted points */
        virtual double operator()(
            const HistoND<Numeric>& histo,
            double effectiveSampleSize,
            const Num2* densityEstimate, unsigned lenEstimate,
            const AbsPolyFilter1D& filterUsed) const = 0;

        /**
        // Cross-validation for samples of univariate weighted points
        // which can be used when the original sample is available.
        // For the sample points, the first element of the pair is
        // the point coordinate and the second element is the weight.
        // The histogram must be created in advance using that particular
        // sample.
        */
        virtual double operator()(
            const HistoND<Numeric>& histo,
            const std::pair<Num3, Num4>* sample, unsigned long lenSample,
            const Num2* densityEstimate, unsigned lenEstimate,
            const AbsPolyFilter1D& filterUsed) const = 0;

        inline double operator()(
            const HistoND<Numeric>& histo,
            const std::vector<std::pair<Num3, Num4> >& sample,
            const Num2* densityEstimate, unsigned lenEstimate,
            const AbsPolyFilter1D& filterUsed) const
        {
            const std::pair<Num3, Num4>* ps = 0;
            const unsigned long lenSample = sample.size();
            if (lenSample)
                ps = &sample[0];
            return this->operator()(histo, ps, lenSample,
                                    densityEstimate, lenEstimate, filterUsed);
        }

#ifdef SWIG
        inline double cv(
            const HistoND<Numeric,HistoAxis>& histo,
            PyObject* input,
            const AbsPolyFilter1D& filterUsed) const
        {
            const int typenum = numpyArrayType(input);
            if (typenum != NumpyTypecode<double>::code)
                throw std::invalid_argument("In npstat::AbsBandwidthCV1D::cv: "
                                            "unsupported input array type");
            const unsigned datalen = histo.nBins();
            if (!isNumpyShapeCompatible(input, &datalen, 1U)) throw std::invalid_argument(
                "In npstat::AbsBandwidthCV1D::cv: incompatible array shape");
            const Num2* in = (double*)PyArray_DATA((PyArrayObject*)input);
            return this->operator()(histo, in, datalen, filterUsed);
        }

        inline double cvWeighted(
            const HistoND<Numeric,HistoAxis>& histo,
            const double effectiveSampleSize,
            PyObject* input,
            const AbsPolyFilter1D& filterUsed) const
        {
            const int typenum = numpyArrayType(input);
            if (typenum != NumpyTypecode<double>::code)
                throw std::invalid_argument("In npstat::AbsBandwidthCV1D::cvWeighted: "
                                            "unsupported input array type");
            const unsigned datalen = histo.nBins();
            if (!isNumpyShapeCompatible(input, &datalen, 1U)) throw std::invalid_argument(
                "In npstat::AbsBandwidthCV1D::cvWeighted: incompatible array shape");
            const Num2* in = (double*)PyArray_DATA((PyArrayObject*)input);
            return this->operator()(histo, effectiveSampleSize, in, datalen, filterUsed);
        }

        inline double cvWeightedSample(
            const HistoND<Numeric,HistoAxis>& histo,
            const std::vector<std::pair<Num3, Num4> >& sample,
            PyObject* input,
            const AbsPolyFilter1D& filterUsed) const
        {
            const int typenum = numpyArrayType(input);
            if (typenum != NumpyTypecode<double>::code)
                throw std::invalid_argument("In npstat::AbsBandwidthCV1D::cvWeightedSample: "
                                            "unsupported input array type");
            const unsigned datalen = histo.nBins();
            if (!isNumpyShapeCompatible(input, &datalen, 1U)) throw std::invalid_argument(
                "In npstat::AbsBandwidthCV1D::cvWeightedSample: incompatible array shape");
            const Num2* in = (double*)PyArray_DATA((PyArrayObject*)input);
            return this->operator()(histo, sample, in, datalen, filterUsed);
        }
#endif // SWIG
    };

    /**
    // Cross-validation for multivariate densities.
    // "Numeric" template parameter is the type of histogram bins.
    // "Array" template parameter should be one of ArrayND types.
    */
    template<typename Numeric, class Array>
    struct AbsBandwidthCVND
    {
        typedef Numeric bin_type;
        typedef Array density_type;

        inline virtual ~AbsBandwidthCVND() {}

        /**
        // It should be assumed that the "nFillsInRange" method of the
        // argument histogram returns the actual number of fills.
        //
        // "densityEstimate" is allowed to be an estimate without truncation
        // (even if it includes negative values).
        */
        virtual double operator()(
            const HistoND<Numeric>& histo,
            const Array& densityEstimate,
            const AbsPolyFilterND& filterUsed) const = 0;

        /** Cross-validation for samples of multivariate weighted points */
        virtual double operator()(
            const HistoND<Numeric>& histo,
            double effectiveSampleSize,
            const Array& densityEstimate,
            const AbsPolyFilterND& filterUsed) const = 0;

#ifdef SWIG
        inline double cv(
            const HistoND<Numeric,HistoAxis>& histo,
            PyObject* input,
            const AbsPolyFilterND& filterUsed) const
        {
            const int typenum = numpyArrayType(input);
            if (typenum != NumpyTypecode<double>::code)
                throw std::invalid_argument("In npstat::AbsBandwidthCVND::cv: "
                                            "unsupported input array type");
            const unsigned* shape = histo.binContents().shapeData();
            if (!isNumpyShapeCompatible(input, shape, histo.dim()))
                throw std::invalid_argument("In npstat::AbsBandwidthCVND::cv: "
                                            "incompatible array shape");
            double* in = (double*)PyArray_DATA((PyArrayObject*)input);
            const ArrayND<double>& wrap = externalMemArrayND(in, shape, histo.dim());
            return this->operator()(histo, wrap, filterUsed);
        }

        inline double cvWeighted(
            const HistoND<Numeric,HistoAxis>& histo,
            const double effectiveSampleSize,
            PyObject* input,
            const AbsPolyFilterND& filterUsed) const
        {
            const int typenum = numpyArrayType(input);
            if (typenum != NumpyTypecode<double>::code)
                throw std::invalid_argument("In npstat::AbsBandwidthCVND::cvWeighted: "
                                            "unsupported input array type");
            const unsigned* shape = histo.binContents().shapeData();
            if (!isNumpyShapeCompatible(input, shape, histo.dim()))
                throw std::invalid_argument("In npstat::AbsBandwidthCVND::cvWeighted: "
                                            "incompatible array shape");
            double* in = (double*)PyArray_DATA((PyArrayObject*)input);
            const ArrayND<double>& wrap = externalMemArrayND(in, shape, histo.dim());
            return this->operator()(histo, effectiveSampleSize, wrap, filterUsed);
        }
#endif // SWIG
    };
}

#endif // NPSTAT_ABSBANDWIDTHCV_HH_
