#include <cmath>
#include <climits>
#include <sstream>
#include <numeric>

#include "npstat/nm/allocators.hh"

#include "npstat/rng/AbsRandomGenerator.hh"

#include "npstat/stat/BinnedADTest1D.hh"
#include "npstat/stat/DiscreteDistributions1D.hh"
#include "npstat/stat/StatUtils.hh"

namespace {
    struct DiscreteADDistance : public npstat::AbsDiscreteDistribution1DDistance
    {
        double operator()(const npstat::AbsDiscreteDistribution1D& data,
                          const npstat::AbsDiscreteDistribution1D& data2,
                          const npstat::AbsDiscreteDistribution1D* pooled,
                          const long first, const long last) const
        {
            const npstat::AbsDiscreteDistribution1D& ref = 
                (pooled ? *pooled : data2);

            long double sum = 0.0L;
            double oldRefCdf = 0.0;
            double oldRefExceedance = 1.0;
            double oldDataCdf = 0.0;
            double oldDataExceedance = 1.0;
            double oldData2Cdf = 0.0;
            double oldData2Exceedance = 1.0;
            bool useExceedance = false;

            double cdfCenterData = 0.0, excCenterData = 0.0;
            double cdfCenterData2 = 0.0, excCenterData2 = 0.0;

            for (long i=first; i<last; ++i)
            {
                const double prob = ref.probability(i);

                const double refCdf = ref.cdf(i);
                const double cdfCenterRef = (refCdf + oldRefCdf)/2.0;
                oldRefCdf = refCdf;

                const double refExceedance = ref.exceedance(i);
                const double excCenterRef = (refExceedance+oldRefExceedance)/2.0;
                oldRefExceedance = refExceedance;

                // Avoid subtractive cancellation by using exceedance
                // for large values of x
                if (useExceedance)
                {
                    const double dataExceedance = data.exceedance(i);
                    excCenterData = (dataExceedance + oldDataExceedance)/2.0;
                    cdfCenterData = 1.0 - excCenterData;
                    oldDataExceedance = dataExceedance;
                    oldDataCdf = 1.0 - oldDataExceedance;

                    if (pooled)
                    {
                        const double data2Exceedance = data2.exceedance(i);
                        excCenterData2=(data2Exceedance+oldData2Exceedance)/2.0;
                        cdfCenterData2 = 1.0 - excCenterData2;
                        oldData2Exceedance = data2Exceedance;
                        oldData2Cdf = 1.0 - oldData2Exceedance;
                    }
                }
                else
                {
                    const double dataCdf = data.cdf(i);
                    cdfCenterData = (dataCdf + oldDataCdf)/2.0;
                    excCenterData = 1.0 - cdfCenterData;
                    oldDataCdf = dataCdf;
                    oldDataExceedance = 1.0 - oldDataCdf;

                    if (pooled)
                    {
                        const double data2Cdf = data2.cdf(i);
                        cdfCenterData2 = (data2Cdf + oldData2Cdf)/2.0;
                        excCenterData2 = 1.0 - cdfCenterData2;
                        oldData2Cdf = data2Cdf;
                        oldData2Exceedance = 1.0 - oldData2Cdf;
                    }

                    useExceedance = (refCdf > 0.5);
                }

                if (prob > 0.0)
                {
                    double delta;
                    if (pooled)
                    {
                        if (useExceedance)
                            delta = excCenterData - excCenterData2;
                        else
                            delta = cdfCenterData - cdfCenterData2;
                    }
                    else
                    {
                        if (useExceedance)
                            delta = excCenterData - excCenterRef;
                        else
                            delta = cdfCenterData - cdfCenterRef;
                    }
                    sum += prob*delta*delta/cdfCenterRef/excCenterRef;
                }
            }
            return sum;
        }
    };
}

namespace npstat {
    BinnedADTest1D::BinnedADTest1D(const bool useTwoSampleTest,
                                   AbsRandomGenerator* rng,
                                   const unsigned mcSamplesForPValue)
        : rng_(rng),
          mcSamples_(mcSamplesForPValue),
          twoSampleTest_(useTwoSampleTest)
    {
    }

    const char* BinnedADTest1D::name() const
    {
        if (name_.empty())
        {
            std::ostringstream os;
            if (twoSampleTest_)
                os << "Two-sample ";
            os << "Anderson-Darling";
            if (rng_ && mcSamples_)
                os << " by MC";
            name_ = os.str();
        }
        return name_.c_str();
    }

    void BinnedADTest1D::compareD(const double* pdata, const double* preference,
                                  const unsigned len, double* distance,
                                  double* pvalue) const
    {
        DiscreteTabulated1D data(0, pdata, len);
        DiscreteTabulated1D ref(0, preference, len);
        DiscreteADDistance calc;
        const double dData = std::accumulate(pdata, pdata+len, 0.0L);
        double dist = 0.0, dRef = 0.0;

        if (twoSampleTest_)
        {
            dRef = std::accumulate(preference, preference+len, 0.0L);
            const DiscreteTabulated1D& pooled = pooledDiscreteTabulated1D(
                data, dData, ref, dRef, 0L, len);
            dist = calc(data, ref, &pooled, 0L, len);
        }
        else
            dist = calc(data, ref, 0, 0L, len);
        if (distance)
            *distance = dist;

        if (pvalue)
        {
            if (rng_ && mcSamples_)
            {
                // Calculate the p-value by MC
                *pvalue = pvalueByPseudo(calc, *rng_, data, dData,
                                         ref, dRef, dist, mcSamples_);
            }
            else
            {
                const double nEff = twoSampleTest_ ? 
                    dData*dRef/(dData + dRef) : dData;
                *pvalue = 1.0 - AD(nEff, nEff*dist);
                if (*pvalue > 1.0)
                    *pvalue = 1.0;
            }
        }
    }
}
