#include <stdexcept>

#include "geners/IOException.hh"
#include "npstat/stat/ComparisonDistribution1D.hh"

namespace npstat {
    double ComparisonDistribution1D::density(const double u) const
    {
        if (u < 0.0 || u > 1.0)
            return 0.0;
        else
        {
            const double x = pH_->quantile(u);
            const double denom = pH_->density(x);
            if (denom == 0.0) throw std::runtime_error(
                "In npstat::ComparisonDistribution1D::density: "
                "baseline density is 0");
            return pF_->density(x)/denom;
        }
    }

    void ComparisonDistribution1D::cleanup()
    {
        delete pH_; pH_ = 0;
        delete pF_; pF_ = 0;
    }

    ComparisonDistribution1D::ComparisonDistribution1D(
        const AbsDistribution1D& compared, const AbsDistribution1D& baseline)
        : AbsDistribution1D(), pF_(0), pH_(0)
    {
        const double xmin = baseline.quantile(0.0);
        const double xmax = baseline.quantile(1.0);
        const double min2 = compared.quantile(0.0);
        const double max2 = compared.quantile(1.0);
        if (!(xmin <= min2 && min2 <= xmax && xmin <= max2 && max2 <= xmax))
            throw std::invalid_argument(
                "In npstat::ComparisonDistribution1D constructor: "
                "incompatible input distributions");
        pF_ = compared.clone();
        pH_ = baseline.clone();
    }

    ComparisonDistribution1D::ComparisonDistribution1D(
        const ComparisonDistribution1D& r)
        : AbsDistribution1D(), pF_(0), pH_(0)
    {
        pF_ = r.pF_->clone();
        pH_ = r.pH_->clone();
    }

    ComparisonDistribution1D& ComparisonDistribution1D::operator=(
        const ComparisonDistribution1D& r)
    {
        if (this != &r)
        {
            AbsDistribution1D::operator=(r);
            cleanup();
            pF_ = r.pF_->clone();
            pH_ = r.pH_->clone();
        }
        return *this;
    }

    bool ComparisonDistribution1D::isEqual(const AbsDistribution1D& other) const
    {
        const ComparisonDistribution1D& r = 
            static_cast<const ComparisonDistribution1D&>(other);
        return *pF_ == *r.pF_ && *pH_ == *r.pH_;
    }

    bool ComparisonDistribution1D::write(std::ostream& os) const
    {
        pF_->classId().write(os);
        pF_->write(os);
        pH_->classId().write(os);
        pH_->write(os);
        return !os.fail();
    }

    ComparisonDistribution1D::ComparisonDistribution1D()
        : pF_(0), pH_(0)
    {
    }

    ComparisonDistribution1D* ComparisonDistribution1D::read(
        const gs::ClassId& id, std::istream& in)
    {
        static const gs::ClassId myClassId(
            gs::ClassId::makeId<ComparisonDistribution1D>());
        myClassId.ensureSameId(id);

        gs::ClassId id1(in, 1);
        AbsDistribution1D* d1 = AbsDistribution1D::read(id1, in);
        gs::ClassId id2(in, 1);
        AbsDistribution1D* d2 = AbsDistribution1D::read(id2, in);
        if (d1 && d2)
        {
            ComparisonDistribution1D* cd = new ComparisonDistribution1D();
            cd->pF_ = d1;
            cd->pH_ = d2;
            return cd;
        }
        else
        {
            delete d1;
            delete d2;
            if (in.fail()) throw gs::IOReadFailure(
                "In npstat::ComparisonDistribution1D::read: "
                "input stream failure");
            else throw gs::IOInvalidData(
                "In npstat::ComparisonDistribution1D::read: "
                "failed to read component distributions");
        }
    }
}
