#include <cmath>
#include <stdexcept>

#include "npstat/stat/DiscreteGauss1DBuilder.hh"

namespace npstat {
    DiscreteGauss1DBuilder::DiscreteGauss1DBuilder(
        const double i_bandwidth, const unsigned dataLen)
        : engine_(2U*dataLen, FFTW_ESTIMATE),
          bufvec_(4U*dataLen),
          bandwidth_(i_bandwidth),
          dataLen_(dataLen)
    {
        if (bandwidth_ < 0.0) throw std::invalid_argument(
            "In npstat::DiscreteGauss1DBuilder constructor: "
            "bandwidth must be non-negative");
        if (bandwidth_ > 0.0)
        {
            const int nPhi = 2U*dataLen_;
            const int nComplex = engine_.imageLength();
            std::vector<std::complex<double> > image(nComplex);
            const double bwsq = bandwidth_*bandwidth_;

            for (int iphi=0; iphi<nPhi; ++iphi)
            {
                const int iy = nPhi - iphi < iphi ? iphi - nPhi : iphi;
                const int icmpl = iphi >= nComplex ? nPhi - iphi : iphi;
                const double cosvm1 = cos(iy*2.0*M_PI/nPhi) - 1.0;
                const double fourier = exp(bwsq*cosvm1);
                image[icmpl] = std::complex<double>(fourier, 0.0);
            }

            engine_.setFilterImage(&image[0], nComplex);
        }
    }

    PolyFilter1D* DiscreteGauss1DBuilder::makeFilter(
        const double* /* taper */, unsigned /* lenTaper */,
        const unsigned binnum, const unsigned datalen) const
    {
        if (datalen != dataLen_) throw std::invalid_argument(
            "In npstat::DiscreteGauss1DBuilder::makeFilter: "
            "inconsistent data length");
        if (binnum >= dataLen_) throw std::out_of_range(
            "In npstat::DiscreteGauss1DBuilder::makeFilter: "
            "bin number out of range");

        PolyFilter1D* filter = new PolyFilter1D(binnum);
        filter->resize(datalen);
        long double* f = &(*filter)[0];

        if (bandwidth_ == 0.0)
        {
            for (unsigned i=0; i<datalen; ++i)
                *f++ = 0.0L;
            (*filter)[binnum] = 1.0L;
        }
        else
        {
            const unsigned nPhi = 2U*dataLen_;
            const unsigned symmPhi = nPhi - binnum - 1U;
            double* in = const_cast<double*>(&bufvec_[0]);
            double* out = in + nPhi;
            in[binnum] = 1.0;
            in[symmPhi] = 1.0;
            (const_cast<ConvolutionEngine1D&>(engine_)).convolveWithFilter(
                in, out, nPhi);
            in[binnum] = 0.0;
            in[symmPhi] = 0.0;
            for (unsigned i=0; i<datalen; ++i)
            {
                const double value = *out++;
                *f++ = value < 0.0 ? 0.0 : value;
            }
        }

        return filter;
    }
}
