#ifndef NPSTAT_DISCRETEGAUSSCOPULASMOOTHER_HH_
#define NPSTAT_DISCRETEGAUSSCOPULASMOOTHER_HH_

/*!
// \file DiscreteGaussCopulaSmoother.hh
//
// \brief Copula smoothing with Green's function for the discrete heat
//        equation using cross-validation
//
// The cross-validation is implemented in the base class, this class only
// needs to build the filters for different bandwidth values.
//
// Author: I. Volobouev
//
// December 2019
*/

#include "npstat/nm/Matrix.hh"

#include "npstat/stat/CVCopulaSmoother.hh"
#include "npstat/stat/SequentialPolyFilterND.hh"

namespace npstat {
    /**
    // This class builds multivariate copula filters which are
    // tensor products of univariate filters
    */
    class DiscreteGaussCopulaSmoother : 
        public CVCopulaSmoother<SequentialPolyFilterND>
    {
    public:
        typedef CVCopulaSmoother<SequentialPolyFilterND> Base;

        /**
        // Constructor arguments are as follows:
        //
        // nBinsInEachDim    -- number of copula bins in each dimension
        //
        // dim               -- copula dimensionality
        //
        // marginTolerance   -- tolerance for the margin to be uniform
        //
        // maxNormCycles     -- max number of copula normalization cycles
        //
        // cvCalc            -- calculator for the quantity being optimized
        //                      in the cross validation process. May be NULL
        //                      in which case cross validation will not
        //                      be used.
        //
        // becomeCvCalcOwner -- tells us whether we should destroy cvCalc
        //                      in our own destructor
        //
        // bandwidthsToUse   -- Bandwidth values to use. The row number of
        //                      the matrix corresponds to the "trial" number
        //                      and the column number gives the bandwidth for
        //                      the corresponding dimension. All bandwidth
        //                      values must be non-negative.
        */
        DiscreteGaussCopulaSmoother(
            const unsigned* nBinsInEachDim, unsigned dim,
            double marginTolerance, unsigned maxNormCycles,
            const CVCalc* cvCalc, bool becomeCvCalcOwner,
            const Matrix<double>& bandwidthsToUse);

        inline virtual ~DiscreteGaussCopulaSmoother() {}
    };
}

#endif // NPSTAT_DISCRETEGAUSSCOPULASMOOTHER_HH_
