#include <cmath>
#include <cassert>
#include <stdexcept>
#include <algorithm>
#include <climits>
#include <cfloat>

#include "npstat/stat/SeriesCGF1D.hh"
#include "npstat/rng/permutation.hh"
#include "npstat/nm/Poly1D.hh"
#include "npstat/nm/findRootUsingBisections.hh"

namespace {
    inline double guessExtraCumulant(const npstat::SeriesCGF1D& cgf,
                                     const double a, const unsigned newOrder)
    {
        assert(newOrder % 2U == 0U);
        assert(newOrder > cgf.maxOrder());
        const double sderi = cgf.derivative(2U, a);
        if (sderi < 0.0)
        {
            const unsigned nm2 = newOrder-2U;
            return -sderi * npstat::ldfactorial(nm2)/powl(a, nm2);
        }
        else
            return 0.0;
    }

    class ConvexificationFcn : public npstat::Functor1<double, double>
    {
    public:
        inline ConvexificationFcn(npstat::SeriesCGF1D& cgf,
                                  const double a, const double b,
                                  const unsigned tune)
            : cgf_(cgf), a_(a), b_(b), tune_(tune) {}

        inline double operator()(const double& cumGuess) const
        {
            cgf_.setCumulant(tune_, cumGuess);
            if (cgf_.isConvex(a_, b_))
                return 1.0;
            else
                return -1.0;
        }

    private:
        npstat::SeriesCGF1D& cgf_;
        double a_;
        double b_;
        unsigned tune_;
    };
}

namespace npstat {
    SeriesCGF1D::SeriesCGF1D(const double* cumulants,
                             const unsigned nCumulants,
                             const double i_smin, const double i_smax)
        : cumulants_(cumulants, cumulants+nCumulants),
          smin_(i_smin),
          smax_(i_smax)
    {
        assert(cumulants);
        if (nCumulants < 2U) throw std::invalid_argument(
            "In npstat::SeriesCGF1D constructor: not enough cumulants");
        if (cumulants[1] <= 0.0) throw std::invalid_argument(
            "In npstat::SeriesCGF1D constructor: variance must be positive");
        if (!(smin_ < 0.0)) throw std::invalid_argument(
            "In npstat::SeriesCGF1D constructor: invalid support infimum");
        if (!(smax_ > 0.0)) throw std::invalid_argument(
            "In npstat::SeriesCGF1D constructor: invalid support supremum");
        truncateLeadingZeros();
    }

    void SeriesCGF1D::setCumulant(const unsigned order, const double value)
    {
        if (order)
        {
            const unsigned maxO = cumulants_.size();
            if (order <= maxO)
            {
                if (order == 2U && value <= 0.0) throw std::invalid_argument(
                    "In npstat::SeriesCGF1D setCumulant: variance must be positive");
                cumulants_[order-1U] = value;
                if (order == maxO && value == 0.0)
                    truncateLeadingZeros();
            }
            else if (value)
            {
                cumulants_.resize(order);
                cumulants_[order-1U] = value;
            }
        }
        else if (value)
        {
            throw std::invalid_argument(
                "In npstat::SeriesCGF1D setCumulant: can't change 0th order cumulant");
        }
    }

    void SeriesCGF1D::truncateLeadingZeros()
    {
        const unsigned sz = cumulants_.size();
        unsigned firstNon0 = sz;
        for (unsigned i=firstNon0-1U; i>0U; --i)
            if (cumulants_[i])
            {
                firstNon0 = i;
                break;
            }
        assert(firstNon0 < sz);
        if (firstNon0 < sz - 1U)
            cumulants_.resize(firstNon0 + 1U);
    }

    bool SeriesCGF1D::isEqual(const AbsCGF1D& other) const
    {
        const SeriesCGF1D& r = static_cast<const SeriesCGF1D&>(other);
        return cumulants_ == r.cumulants_ && smin_ == r.smin_ && smax_ == r.smax_;
    }

    double SeriesCGF1D::derivative(const unsigned order, const double s) const
    {
        if (s == 0.0)
        {
            if (order)
            {
                const unsigned idx = order - 1U;
                if (idx < cumulants_.size())
                    return cumulants_[idx];
            }
        }
        else if (!(smin_ < s && s < smax_)) throw std::invalid_argument(
            "In npstat::SeriesCGF1D::derivative: argument out of range");
        else
        {
            const unsigned degree = cumulants_.size();
            if (order <= degree)
            {
                const double* a = &cumulants_[0];
                long double sum = 0.0L;
                if (order)
                {
                    const int shift = order-1U;
                    for (int deg=degree-order; deg>=0; --deg)
                    {
                        sum *= s;
                        sum += a[deg+shift]/ldfactorial(deg);
                    }
                }
                else
                {
                    for (unsigned deg=degree; deg>0U; --deg)
                    {
                        sum += a[deg-1]/ldfactorial(deg);
                        sum *= s;
                    }
                }
                return sum;
            }
        }
        return 0.0;
    }

    SeriesCGF1D* SeriesCGF1D::shiftAndScale(const double mu,
                                            const double sigma) const
    {
        if (!(sigma > 0.0)) throw std::invalid_argument(
            "In npstat::SeriesCGF1D::shiftAndScale: scale must be positive");
        std::vector<double> cum(cumulants_);
        const unsigned sz = cum.size();
        cum[0] -= mu;
        if (sigma != 1.0)
            for (unsigned i=0; i<sz; ++i)
                cum[i] /= pow(sigma, i+1U);

        // Rethink this. It is not obvious how the
        // support should be modified.
        return new SeriesCGF1D(&cum[0], sz, smin_, smax_);
    }

    bool SeriesCGF1D::isConvex(const double a, const double b) const
    {
        if (a >= b) throw std::invalid_argument(
            "In npstat::SeriesCGF1D::sisConvex: invalid interval");
        const unsigned degree = cumulants_.size();
        std::vector<long double> coeffs(degree + 1U);
        coeffs[0] = 0.0L;
        for (unsigned deg=0; deg<degree; ++deg)
            coeffs[deg+1] = cumulants_[deg]/ldfactorial(deg+1U);
        const Poly1D cgf0(coeffs);
        const Poly1D& cgf1(cgf0.derivative());
        const Poly1D& cgf2(cgf1.derivative());
        return cgf2((a + b)/2.0) > 0.0L && !cgf2.nRoots(a, b);
    }

    SeriesCGF1D SeriesCGF1D::convexify(const double a, const double b) const
    {
        if (isConvex(a, b))
            return *this;

        const unsigned maxO = cumulants_.size();
        const unsigned tune = maxO % 2U ? maxO + 1U : maxO + 2U;
        const double cumGuessA = guessExtraCumulant(*this, a, tune);
        const double cumGuessB = guessExtraCumulant(*this, b, tune);
        double cumGuess = std::max(cumGuessA, cumGuessB);
        SeriesCGF1D copy(*this);
        if (cumGuess > 0.0)
        {
            copy.setCumulant(tune, cumGuess);
            const double step = (b - a)*sqrt(DBL_EPSILON);
            if (copy.isConvex(a + step, b - step))
                return copy;
        }

        cumGuess = std::max(cumGuess, 1.0);
        copy.setCumulant(tune, cumGuess);
        while (!copy.isConvex(a, b))
        {
            assert(cumGuess < DBL_MAX/2.0);
            cumGuess *= 2.0;
            copy.setCumulant(tune, cumGuess);
        }

        ConvexificationFcn fcn(copy, a, b, tune);
        double cum = 0.0;
        const double tol = cumGuess*DBL_EPSILON*2.0;
        const bool status = findRootUsingBisections(
            fcn, 0.0, 0.0, cumGuess, tol, &cum);
        assert(status);
        copy.setCumulant(tune, cum);
        return copy;
    }
}
